/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <assert.h>


#include "traffic_model.h"

/****************** QUICK SORT *************************/
void quick_sort(double *data_list, int start, int end)
{
  register uint32_t i, k;
  register double d;
  if (start >= end)
    return;

  i = start;
  d = data_list[start];
  for (k = start+1; k <= end; k++)
    if (data_list[k] < d) {
      data_list[i] = data_list[k];
      data_list[k] = data_list[i+1];
      i++;
    }      
  data_list[i] = d;
  quick_sort(data_list, start, i-1);
  quick_sort(data_list, i+1, end);
}

inline double data_mean(double *data_list, int data_counter)
{
  register int i;
  register double d;
  
  d = 0;
  for (i = 0; i < data_counter; i++)
    d += data_list[i];

  return d/data_counter;
}

inline double data_variance(double *data_list, int data_counter, double mean)
{
  register int i;
  register double d;

  d = 0;
  for (i = 0; i < data_counter; i++)
    d += (data_list[i]-mean)*(data_list[i]-mean);

  return d/data_counter;
}

void pareto_model(double *data_list, int data_counter)
{
  double mean, var, a, b;
  FILE *outfile;
  int i;

  if (data_counter <= 0)
    return;

  quick_sort(data_list, 0, data_counter-1);

  outfile = fopen("gplot-par.dat", "w");
  for (i = 0; i < data_counter; i++)
    fprintf(outfile, "%f\t%f\n", data_list[i], (i+1.0)/data_counter);
  fclose(outfile);

  mean = data_mean(data_list, data_counter);
  var = data_variance(data_list, data_counter, mean);
  printf("\tmean=%f\tvar=%f\trecord_num=%u\n", mean, var, data_counter);

  if (var == 0) {
    printf("\tConst(%f)\tmax=%f\n", mean, data_list[data_counter-1]);
    return;
  }

  b = sqrt(mean*mean/var+1)+1;
  a = (b-1)*(b-2)*var/mean;
  printf("\tPareto(%f, %f)\tmax=%f\n", a, b, data_list[data_counter-1]);

  outfile = fopen("gplot-par.plt", "w");
  fprintf(outfile, "set xrange [0:%f]\n", data_list[data_counter-1]*2);
  fprintf(outfile, "set yrange [0:1]\n");
  //fprintf(outfile, "set logscale x\n");
  fprintf(outfile, "set grid\n");
  fprintf(outfile, "plot \"gplot-par.dat\" using 1:2 title \"Real data\" with points\n");
  fprintf(outfile, "a=%f\n", a);
  fprintf(outfile, "b=%f\n", b);
  fprintf(outfile, "f(x)=1-(a/x)**b\n");
  fprintf(outfile, "replot 1-(%f/x)**%f title \"Pareto(%f, %f)\" with lines lw 2\n", a, b, a, b);
  fprintf(outfile, "fit f(x) \"gplot-par.dat\" using 1:2 via a, b\n");
  fprintf(outfile, "replot f(x) title \"Gnuplot modeling\"\n");
  fprintf(outfile, "print \"\\nPareto(\", a, \",\", b, \")\"");
  fclose(outfile);
}

void exponential_model(double *data_list, int data_counter)
{
  double mean, var, lambda;
  FILE *outfile;
  int i;

  if (data_counter <= 0)
    return;

  quick_sort(data_list, 0, data_counter-1);

  mean = data_mean(data_list, data_counter);
  var = data_variance(data_list, data_counter, mean);
  lambda = 1.0 / mean;

  printf("\tmean=%f\tvar=%f\trecord_num=%u\n", mean, var, data_counter);
  printf("\tExponential(%f)\tmax=%f\n", lambda, data_list[data_counter-1]);

  outfile = fopen("gplot-exp.dat", "w");
  for (i = 0; i < data_counter; i++)
    fprintf(outfile, "%f\t%f\n", data_list[i], (i+1.0)/data_counter);
  fclose(outfile);

  outfile = fopen("gplot-exp.plt", "w");
  fprintf(outfile, "set xrange [0:%f]\n", data_list[data_counter-1]*2);
  fprintf(outfile, "set yrange [0:1]\n");
  fprintf(outfile, "set grid\n");
  fprintf(outfile, "plot \"gplot-exp.dat\" using 1:2 title \"Real data\" with points\n");
  fprintf(outfile, "a=%f\n", lambda);
  fprintf(outfile, "f(x)=1-exp(-a*x)\n");
  fprintf(outfile, "replot 1-exp(-%f*x) title \"Exponential(%f)\" with lines lw 2\n", lambda, lambda);
  fprintf(outfile, "fit f(x) \"gplot-exp.dat\" using 1:2 via a\n");
  fprintf(outfile, "replot f(x) title \"Gnuplot modeling\"\n");
  fprintf(outfile, "print \"\\nExponetial(\", a,  \")\"");
  fclose(outfile);
}

void poisson_model(double *data_list, int data_counter)
{
  double mean, var, p, L;
  FILE *outfile;
  uint32_t i, k;
  double *temp_list;

  if (data_counter <= 0)
    return;

  quick_sort(data_list, 0, data_counter);

  mean = data_mean(data_list, data_counter);
  var = data_variance(data_list, data_counter, mean);

  printf("\tmean=%f\tvar=%f\trecord_num=%u\n", mean, var, data_counter);
  printf("\tPoisson(%f)\tmax=%f\n", mean, data_list[data_counter-1]);

  
  temp_list = malloc(sizeof(double)*data_counter);
  L = exp(-mean);
  srand(time(NULL));
  for (i = 0; i < data_counter; i++) {
    k = 0;
    p = 1;
    do {
      k++;
      p *= rand() * 1.0 / RAND_MAX;
    } while (p >= L);
    temp_list[i] = k - 1;
  }
  quick_sort(temp_list, 0, data_counter-1);
  

  outfile = fopen("gplot-poi.dat", "w");
  if (outfile == NULL) {
    printf("Failed to open the file gplot.dat\n");
    exit(-1);
  }
  for (i = 0; i < data_counter; i++) {
    fprintf(outfile, "%f\t%f\t%f\n", data_list[i], temp_list[i], (i+1.0)/data_counter);
  }
  fclose(outfile);

  outfile = fopen("gplot-poi.plt", "w");
  fprintf(outfile, "set xrange [0:%f]\n", temp_list[data_counter-1]);
  fprintf(outfile, "set yrange [0:1]\n");
  fprintf(outfile, "set grid\n");
  fprintf(outfile, "plot \"gplot-poi.dat\" using 1:3 title \"Real data\" with points\n");
  //fprintf(outfile, "a=%f\n", mean);
  //fprintf(outfile, "f(x)=(gamma(int(x+1))-igamma(int(x+1),abs(a)))/gamma(int(x)+1)\n");
  //fprintf(outfile, "replot f(x)\n");
  fprintf(outfile, "replot \"gplot-poi.dat\" using 2:3 title \"Poisson(%f)\" with lines\n", mean);
  fclose(outfile);
}


/***************** HASHTABLE FUNCTIONS *****************/

/* compares two hashtable entries */
int compare_server_stats(const void * entry1, const void * entry2)
{
  const server_stats *foo1 = entry1;
  const server_stats *foo2 = entry2;
  
  return (foo1->ip_address != foo2->ip_address);
}
int compare_client_stats(const void * entry1, const void * entry2)
{
  const client_stats *foo1 = entry1;
  const client_stats *foo2 = entry2;
  
  return (foo1->ip_address != foo2->ip_address || foo1->parent != foo2->parent);
}
int compare_communication_stats(const void * entry1, const void * entry2)
{
  const communication_stats *foo1 = entry1;
  const communication_stats *foo2 = entry2;
  
  return (foo1->port_number != foo2->port_number || foo1->parent != foo2->parent);
}

/* deletes a hashtable entry */
void delete_server_stats(void * entry)
{
  server_stats *temp = entry;
  if (!temp) return;
  free(temp);
  server_hash_count--;
}
void delete_client_stats(void * entry)
{
  client_stats *temp = entry;
  if (!temp) return;
  free(temp);
  client_hash_count--;
}
void delete_communication_stats(void * entry)
{
  communication_stats *temp = entry;
  if (!temp) return;
  free(temp);
  communication_hash_count--;
}

/* makes a hashtable entry */
static unsigned long make_key_server_stats(const void * entry)
{
  const server_stats *temp = entry;
  
  return temp->ip_address;
}
static unsigned long make_key_client_stats(const void * entry)
{
  const client_stats *temp = entry;
  
  return (unsigned long)temp->ip_address + (unsigned long)temp->parent->ip_address;
}
static unsigned long make_key_communication_stats(const void * entry)
{
  const communication_stats *temp = entry;
  
  return (unsigned long)temp->parent + (unsigned long)temp->port_number;
}

static void init_hashes(void)
{
  server_hash = init_hash_table("server_stats hash", compare_server_stats, make_key_server_stats, delete_server_stats, SERVER_STATS_HASH_SIZE);
  client_hash = init_hash_table("client_stats hash", compare_client_stats, make_key_client_stats, delete_client_stats, CLIENT_STATS_HASH_SIZE);
  communication_hash = init_hash_table("communication_stats hash", compare_communication_stats, make_key_communication_stats, delete_communication_stats, COMMUNICATION_STATS_HASH_SIZE);

  server_hash_count = 0;
  client_hash_count = 0;
  communication_hash_count = 0;
}


/***************BEGIN PROCESSING REQ/REPLY ************/
void process_request(struct server_stats *server_ptr, struct ip *ip, void *ip_payload, struct timespec *ts) {
  struct tcphdr *tcp;
  struct udphdr *udp;
  struct client_stats *client_ptr, tmp_client;
  struct communication_stats *communication_ptr, tmp_communication;
  unsigned int data_len;

  tmp_client.parent = server_ptr;
  tmp_client.ip_address = ntohl(ip->ip_src.s_addr);
  client_ptr = (struct client_stats*)find_hash_entry(client_hash, &tmp_client);
  if (client_ptr == NULL) {
    client_ptr = malloc(sizeof(struct client_stats));
    client_ptr->ip_address = tmp_client.ip_address;
    client_ptr->parent = server_ptr;
    client_ptr->request_count = 0;
    client_ptr->com_list = NULL;
    client_ptr->next = server_ptr->clients;
    server_ptr->clients = client_ptr;
    server_ptr->client_count++;
    add_hash_entry(client_hash, client_ptr);
    client_hash_count++;
  }

  if (ip->ip_p == IPPROTO_TCP) {
    tcp = (struct tcphdr*)ip_payload;

    //return if there is no payload in the TCP packet
    if (ntohs(ip->ip_len) - (ip->ip_hl + tcp->th_off) * 4 <= 1)
      return;

    tmp_communication.parent = client_ptr;
    tmp_communication.port_number = ntohs(tcp->th_sport);
    communication_ptr = (struct communication_stats*)find_hash_entry(communication_hash, &tmp_communication);

    if (communication_ptr != NULL && communication_ptr->communication_finalized == 1) {
      communication_ptr->port_number += 65536*2;
      communication_ptr->communication_finalized = 2;
      communication_ptr = NULL;
    }

    if (communication_ptr == NULL) {
      communication_ptr = malloc(sizeof(struct communication_stats));
      communication_ptr->request_length = 0;
      communication_ptr->response_length = 0;
      communication_ptr->parent = client_ptr;
      communication_ptr->port_number = ntohs(tcp->th_sport);
      communication_ptr->tcp_seq = 0;
      communication_ptr->time_stamp = *ts;
      communication_ptr->communication_finalized = 0;
      communication_ptr->next = client_ptr->com_list;
      client_ptr->com_list = communication_ptr;
      client_ptr->request_count++;
      client_ptr->parent->request_count++;
      add_hash_entry(communication_hash, communication_ptr);
      communication_hash_count++;
    }

    data_len = ntohs(ip->ip_len) - (ip->ip_hl + tcp->th_off) * 4;
    if (communication_ptr->tcp_seq == 0)
      communication_ptr->tcp_seq = ntohl(tcp->th_seq);
    if ( ntohl(tcp->th_seq) + data_len > communication_ptr->tcp_seq) {
      //printf("%u %u %u\n", communication_ptr->tcp_seq, ntohl(tcp->th_seq) + data_len, data_len);
      communication_ptr->request_length += ntohl(tcp->th_seq) + data_len - communication_ptr->tcp_seq;
      //printf("%u %u %u %u\n", communication_ptr->tcp_seq, ntohl(tcp->th_seq) + data_len, data_len, communication_ptr->request_length);
      communication_ptr->tcp_seq = ntohl(tcp->th_seq) + data_len;
    }
  }
  else if (ip->ip_p == IPPROTO_UDP) {
    udp = (struct udphdr*)ip_payload;

    tmp_communication.parent = client_ptr;
    tmp_communication.port_number = ntohs(udp->uh_sport) + 65536;
    communication_ptr = (struct communication_stats*)find_hash_entry(communication_hash, &tmp_communication);

    if (communication_ptr != NULL && communication_ptr->communication_finalized == 1) {
      communication_ptr->port_number += 65536*2;
      communication_ptr->communication_finalized = 2;
      communication_ptr = NULL;
    }

    if (communication_ptr == NULL) {
      communication_ptr = malloc(sizeof(struct communication_stats));
      communication_ptr->request_length = 0;
      communication_ptr->response_length = 0;
      communication_ptr->parent = client_ptr;
      communication_ptr->port_number = ntohs(udp->uh_sport) + 65536;
      communication_ptr->time_stamp = *ts;
      communication_ptr->communication_finalized = 0;
      communication_ptr->next = client_ptr->com_list;
      client_ptr->com_list = communication_ptr;
      client_ptr->request_count++;
      client_ptr->parent->request_count++;
      add_hash_entry(communication_hash, communication_ptr);
      communication_hash_count++;      
    }

    assert(ntohs(udp->uh_ulen) >= 8);
    communication_ptr->request_length += ntohs(udp->uh_ulen) - 8;
  }
  
  if (current_time == 0)
    current_time = ts->tv_sec;
  else if (current_time < ts->tv_sec) {
    clock_second_counter += ts->tv_sec - current_time;
    current_time = ts->tv_sec;
  }
  assert(clock_second_counter < TRACE_EXPECTED_DURATION);
  req_per_second[clock_second_counter]++;
}

void process_response(struct server_stats *server_ptr, struct ip *ip, void *ip_payload, struct timespec *ts) {
  struct tcphdr *tcp;
  struct udphdr *udp;
  struct client_stats *client_ptr, tmp_client;
  struct communication_stats *communication_ptr, tmp_communication;;
  unsigned int data_len;

  tmp_client.parent = server_ptr;
  tmp_client.ip_address = ntohl(ip->ip_dst.s_addr);
  client_ptr = (struct client_stats*)find_hash_entry(client_hash, &tmp_client);
  if (client_ptr == NULL) {
    client_ptr = malloc(sizeof(struct client_stats));
    client_ptr->ip_address = tmp_client.ip_address;
    client_ptr->parent = server_ptr;
    client_ptr->request_count = 0;
    client_ptr->com_list = NULL;
    client_ptr->next = server_ptr->clients;
    server_ptr->clients = client_ptr;
    server_ptr->client_count++;
    add_hash_entry(client_hash, client_ptr);
    client_hash_count++;
  }

  if (ip->ip_p == IPPROTO_TCP) {
    tcp = (struct tcphdr*)ip_payload;
    data_len = ntohs(ip->ip_len) - (ip->ip_hl + tcp->th_off) * 4;
    if (data_len < 1)
      return;

    tmp_communication.parent = client_ptr;
    tmp_communication.port_number = ntohs(tcp->th_dport);
    communication_ptr = (struct communication_stats*)find_hash_entry(communication_hash, &tmp_communication);

    if (communication_ptr != NULL && communication_ptr->communication_finalized < 2) {
      if (communication_ptr ->communication_finalized < 1) {
	communication_ptr->communication_finalized = 1;
	communication_ptr->tcp_seq = ntohl(tcp->th_seq);
      }
      if (ntohl(tcp->th_seq) + data_len > communication_ptr->tcp_seq) {
	communication_ptr->response_length += ntohl(tcp->th_seq) + data_len - communication_ptr->tcp_seq;
	communication_ptr->tcp_seq = ntohl(tcp->th_seq) + data_len;
      }
    }
  }
  else if (ip->ip_p == IPPROTO_UDP) {
    udp = (struct udphdr*)ip_payload;

    tmp_communication.parent = client_ptr;
    tmp_communication.port_number = ntohs(udp->uh_dport) + 65536;
    communication_ptr = (struct communication_stats*)find_hash_entry(communication_hash, &tmp_communication);

    if (communication_ptr != NULL) {
      communication_ptr->communication_finalized = 1;
      assert(ntohs(udp->uh_ulen) >= 8);
      communication_ptr->response_length += ntohs(udp->uh_ulen) - 8;
    }
    else {
      //we have not seen a corresponding request for this reply
    }
  }
}

/*** MAIN ***/
static void quit(int sig)
{
  coral_pkt_done = 1;
}

void add_to_data_list(double d)
{
  assert(d > 0);

  if (data_list_capacity <= data_list_counter) {
    if (data_list_capacity == 0) {
      data_list_capacity = INITIAL_DATA_LIST_SIZE;
      data_list = malloc(sizeof(double) * data_list_capacity);
    }
    else {
      data_list_capacity *= 2;
      data_list = realloc(data_list, sizeof(double) * data_list_capacity);
    }
  }

  data_list[data_list_counter] = d;
  data_list_counter++;
}

int main (int argc, char *argv[])
{
  int i, j, n;

  coral_source_t *sources;
  char source_file[MAX_SOURCES][256];
  int trace_file_counter;
  coral_iface_t *iface;
  coral_pkt_result_t pkt_result;
  coral_interval_result_t interval_result;
  coral_pkt_buffer_t netpkt, payload;

  struct ip *ip;
  struct tcphdr *tcp;
  struct udphdr *udp;

  uint64_t total_packet = 0;
  uint64_t total_byte = 0;

  struct timeval interval = {3600, 0};
  struct timespec ts;
  uint32_t last_request_time;
  uint32_t same_second_request_counter;

  uint32_t src_addr, dst_addr;
  uint32_t src_port, dst_port;
  uint32_t monitored_src_subnet, monitored_dst_subnet, monitored_port;
  uint32_t src_subnet_mask_length, dst_subnet_mask_length;

  struct server_stats tmp_server, *server_ptr;
  struct client_stats tmp_client, *client_ptr;
  struct communication_stats tmp_communication, *communication_ptr;

  uint32_t i1, i2, i3, i4, i5;

  //to count the number of active address in subnet
  uint32_t *active_host;
  double *interval_active_host;
  uint32_t time_interval_counter;

  //to measure the life of an address
  uint32_t *host_start_time;
  uint32_t *host_end_time;


  signal(SIGINT, quit);

  //Print out instructions
  if (argc < 6) {
    printf("\tUsage: %s server-net client-net port stat tracefile ....\n", argv[0]);
    printf("\t\tstat:\t1  request inter-arrival time\n");
    printf("\t\t\t2  request size\n");
    printf("\t\t\t3  resply size\n");
    printf("\t\t\t4  request per host\n");
    printf("\t\t\t5  active host count\n");
    printf("\t\t\t6  host active time\n");
    printf("\texample: %s 0.3.117.0/24 225.113.32.0/24 tcp80 1 twoway.pcap\n", argv[0]);
    printf("\texample: %s 0.3.117.0/16 225.113.32.0/24 udp53 2 twoway.pcap\n", argv[0]);
    return 1;      
  }


  //Initialize CoralReef API
  coral_set_api(CORAL_API_PKT);
  coral_set_options(0, CORAL_OPT_SORT_TIME | CORAL_OPT_RAW_TIME);
  coral_set_iomode(0, CORAL_RX_UNKNOWN, -1, 0); 
  coral_set_duration(0);

  //reading the arguments
  printf("Processing ");
  sscanf(argv[1], "%u.%u.%u.%u/%u (servers)", &i1, &i2, &i3, &i4, &i5);
  monitored_src_subnet = ((i1 << 24) + (i2 << 16) + (i3 << 8) + i4) & (0xFFFFFFFF << (32-i5));
  src_subnet_mask_length = i5;
  printf("%u.%u.%u.%u/%u <-> ", monitored_src_subnet>>24&0xFF, monitored_src_subnet>>16&0xFF, monitored_src_subnet>>8&0xFF, monitored_src_subnet&0xFF, i5);

  sscanf(argv[2], "%u.%u.%u.%u/%u (clients)", &i1, &i2, &i3, &i4, &i5);
  monitored_dst_subnet = ((i1 << 24) + (i2 << 16) + (i3 << 8) + i4) & (0xFFFFFFFF << (32-i5));
  dst_subnet_mask_length = i5;
  printf("%u.%u.%u.%u/%u ", monitored_dst_subnet>>24&0xFF,  monitored_dst_subnet>>16&0xFF, monitored_dst_subnet>>8&0xFF, monitored_dst_subnet&0xFF, i5);

  printf(" at port ");

  //monitored service port
  if (strstr(argv[3], "tcp") != NULL || strstr(argv[3], "TCP") != NULL) {
    monitored_port = atoi(argv[3] + 3);
    printf("TCP%u\n", monitored_port);
  }
  else if (strstr(argv[3], "udp") != NULL || strstr(argv[3], "UDP") != NULL) {
    monitored_port = atoi(argv[3] + 3) + 65536;
    printf("UDP%u\n", monitored_port - 65536);
  }
  else {
    printf("Invalid argument: \"%s\"\n", argv[3]);
    exit(-1);
  }
  
  
  //Initialize statistics
  req_per_second = (uint32_t*)malloc(sizeof(uint32_t) * TRACE_EXPECTED_DURATION);
  memset(req_per_second, 0, sizeof(uint32_t) * TRACE_EXPECTED_DURATION);
  current_time = 0;
  clock_second_counter = 0;


  //Initialize the hashtables
  init_hashes();


  //Read and store the trace-file names
  optind = 5;
  trace_file_counter = 0;
  while (optind < argc) {
    strcpy(source_file[trace_file_counter], argv[optind]);
    optind++;
    trace_file_counter++;
  }


  active_host = (uint32_t*)malloc(sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));
  memset(active_host, 0, sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));

  interval_active_host = (double*)malloc(sizeof(double) * (TRACE_EXPECTED_DURATION / TRACE_TIME_INTERVAL));
  memset(interval_active_host, 0, sizeof(double) * (TRACE_EXPECTED_DURATION / TRACE_TIME_INTERVAL));
  time_interval_counter = 0;

  host_start_time = (uint32_t*)malloc(sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));
  memset(host_start_time, 0, sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));

  host_end_time = (uint32_t*)malloc(sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));
  memset(host_end_time, 0, sizeof(uint32_t) * (1 << (32-src_subnet_mask_length)));

  
  //Processing all the trace files
  for (i = 0; i < trace_file_counter; i++) {
    if ((sources = coral_new_source(source_file[i])) == NULL)
      exit(-1);
    n = coral_open(sources);
    if (n == 0) {
      coral_diag(0, ("Could not open sources.\n"));
    }
    if (coral_start(sources) < 0) {
      fprintf(stderr, "Could not start source %s, aborting!\n", source_file[i]);
      exit(-1);
    }

    coral_read_pkt_init(NULL, NULL, &interval);

    while (1) {
      if (coral_pkt_done) {
        coral_stop_all();
        exit(-1);
      }

      iface = coral_read_pkt(&pkt_result, &interval_result);

      // no more packets 
      if (!iface)
        break;
      
      if (pkt_result.packet) {
	CORAL_TIMESTAMP_TO_TIMESPEC(iface, pkt_result.timestamp, &ts);
	
	if (coral_get_payload_by_layer(pkt_result.packet, &netpkt, 3) < 0) {
	  continue;
	}
	if (netpkt.protocol != CORAL_NETPROTO_IP) {
	  continue;
	}
	
	total_packet++;
	
	ip = (struct ip*)netpkt.buf;
	src_addr = ntohl(ip->ip_src.s_addr);
	dst_addr = ntohl(ip->ip_dst.s_addr);
		
	coral_get_payload(&netpkt, &payload);
	if (ip->ip_p == IPPROTO_TCP) {
	  tcp = (struct tcphdr*)payload.buf;
	  src_port = ntohs(tcp->th_sport);
	  dst_port = ntohs(tcp->th_dport);
	}
	else if (ip->ip_p == IPPROTO_UDP) {
	  udp = (struct udphdr*)payload.buf;
	  src_port = ntohs(udp->uh_sport) + 65536;
	  dst_port = ntohs(udp->uh_dport) + 65536;
	}
	else
	  continue;

	if ((src_addr & (0xFFFFFFFF<<(32 - src_subnet_mask_length))) == monitored_src_subnet
	    && (dst_addr & (0xFFFFFFFF<<(32 - dst_subnet_mask_length))) == monitored_dst_subnet
	    && src_port == monitored_port) {
	  tmp_server.ip_address = src_addr;
	  if ((server_ptr = (server_stats*)find_hash_entry(server_hash, &tmp_server)) == NULL) {
	    server_ptr = malloc(sizeof(struct server_stats));
	    server_ptr->ip_address = src_addr;
	    server_ptr->request_count = 0;
	    server_ptr->client_count = 0;
	    server_ptr->clients = NULL;
	    add_hash_entry(server_hash, server_ptr);
	    server_hash_count++;
	  }
	  process_response(server_ptr, ip, (void*)payload.buf, &ts);
	}
	else if ((dst_addr & (0xFFFFFFFF<<(32 - src_subnet_mask_length))) == monitored_src_subnet 
		 && (src_addr & (0xFFFFFFFF<<(32 - dst_subnet_mask_length))) == monitored_dst_subnet 
		 && dst_port == monitored_port) {
	  tmp_server.ip_address = dst_addr;
	  if ((server_ptr = (server_stats*)find_hash_entry(server_hash, &tmp_server)) == NULL) {
	    server_ptr = malloc(sizeof(struct server_stats));
	    server_ptr->ip_address = dst_addr;
	    server_ptr->request_count = 0;
	    server_ptr->client_count = 0;
	    server_ptr->clients = NULL;
	    add_hash_entry(server_hash, server_ptr);
	    server_hash_count++;
	  }
	  process_request(server_ptr, ip, (void*)payload.buf, &ts);
	}


	//counting number of active hosts in 5-minute-interval
	if ((dst_addr & (0xFFFFFFFF<<(32 - dst_subnet_mask_length))) == monitored_dst_subnet 
	    && (src_addr & (0xFFFFFFFF<<(32 - src_subnet_mask_length))) == monitored_src_subnet) {

	  active_host[src_addr & (0xFFFFFFFF >> src_subnet_mask_length)] = 1;

	  if (host_start_time[src_addr & (0xFFFFFFFF >> src_subnet_mask_length)] == 0)
	    host_start_time[src_addr & (0xFFFFFFFF >> src_subnet_mask_length)] = current_time;
	  host_end_time[src_addr & (0xFFFFFFFF >> src_subnet_mask_length)] = current_time;
	}

	if (time_interval_counter * TRACE_TIME_INTERVAL < clock_second_counter) {
	  for (n = 0; n < (1 << (32 - src_subnet_mask_length)); n++)
	    if (active_host[n] > 0)
	      interval_active_host[time_interval_counter]++;
	  time_interval_counter++;
	       memset(active_host, 0, sizeof(uint32_t) * (1 << (32 - src_subnet_mask_length)));
	}
      }
      else if (!interval_result.stats) {
	//process interval result
      }
      else {
	//no packet or interval data to process
      }
    }
    coral_stop(sources);
    coral_close(sources);
  }
  
  
  /*
  //print out the original statistics (raw data)

  //printf("Total packet: %"PRIu64"\tTotal byte: %"PRIu64"\n", total_packet, total_byte);
  //printf("Total_server=%i\tTotal_client=%i\tTotal_communication=%i\n", server_hash_count, client_hash_count, communication_hash_count);

  init_hash_walk(server_hash);
  while ((server_ptr = next_hash_walk(server_hash)) != NULL) {
    src_addr = htonl(server_ptr->ip_address);
    printf("%s\t", inet_ntoa(*((struct in_addr*)&src_addr)));
    printf("request_count = %u\tclient_count = %u\n", server_ptr->request_count, server_ptr->client_count);

    client_ptr = server_ptr->clients;
    while (client_ptr != NULL) {
      if (client_ptr->request_count > 0) {
	src_addr = htonl(client_ptr->ip_address);
	printf("\t%s\t%u", inet_ntoa(*((struct in_addr*)&src_addr)), client_ptr->request_count);
	communication_ptr = client_ptr->com_list;
	while (communication_ptr != NULL) {
	  if (communication_ptr->request_length > 0 && communication_ptr->response_length > 0)
	    printf(" <%u %u %u>", communication_ptr->request_length, communication_ptr->response_length, communication_ptr->time_stamp.tv_sec);
	  assert(communication_ptr != communication_ptr->next);
	  communication_ptr = communication_ptr->next;
	}
	printf("\n");
      }
      client_ptr = client_ptr->next;
    }
  }
  */

  data_list_counter = 0;
  data_list_capacity = 0;
  data_list = NULL;

  if (strcmp(argv[4], "1") == 0) {
    printf("Distribution of request inter-arrival time:\n");
    init_hash_walk(server_hash);
    while ((server_ptr = next_hash_walk(server_hash)) != NULL) {
      client_ptr = server_ptr->clients;
      while (client_ptr != NULL) {
	communication_ptr = client_ptr->com_list;
	last_request_time = 0;
	same_second_request_counter = 0;
	while (communication_ptr != NULL) {
	  if (communication_ptr->request_length > 0 && communication_ptr->response_length > 0) {
	    if (last_request_time <= 0)
	      last_request_time = communication_ptr->time_stamp.tv_sec;
	    else if (last_request_time == communication_ptr->time_stamp.tv_sec)
	      same_second_request_counter++;
	    else if (last_request_time > communication_ptr->time_stamp.tv_sec) {
	      if (same_second_request_counter > 0) {
		for (i = 0; i < same_second_request_counter; i++)
		  add_to_data_list(1.0/same_second_request_counter);
		same_second_request_counter = 0;
	      }
	      add_to_data_list(last_request_time - communication_ptr->time_stamp.tv_sec);
	      last_request_time = communication_ptr->time_stamp.tv_sec;
	    }
	    else {
	      src_addr = htonl(server_ptr->ip_address);
	      printf("Invalid timestamp sequence at %s -> ", inet_ntoa(*((struct in_addr*)&src_addr)));
	      src_addr = htonl(client_ptr->ip_address);
	      printf("%s\n", inet_ntoa(*((struct in_addr*)&src_addr)));
	      exit(1);
	    }
	  }
	  communication_ptr = communication_ptr->next;
	}
	client_ptr = client_ptr->next;
      }
    }
    
    quick_sort(data_list, 0, data_list_counter-1);
    data_list_counter = data_list_counter * 0.9; //to remove the extrumely abnormal data
    exponential_model(data_list, data_list_counter);
  }
  else if (strcmp(argv[4], "2") == 0) {
    printf("Distribution of request size:\n");
    init_hash_walk(server_hash);
    while ((server_ptr = next_hash_walk(server_hash)) != NULL) {
      client_ptr = server_ptr->clients;
      while (client_ptr != NULL) {
	communication_ptr = client_ptr->com_list;
	while (communication_ptr != NULL) {
	  if (communication_ptr->request_length > 0 && communication_ptr->response_length > 0)
	    add_to_data_list(communication_ptr->request_length);
	  communication_ptr = communication_ptr->next;
	}
	client_ptr = client_ptr->next;
      }
    }
    pareto_model(data_list, data_list_counter);
  }
  else if (strcmp(argv[4], "3") == 0) {
    printf("Distribution of reply size:\n");
    init_hash_walk(server_hash);
    while ((server_ptr = next_hash_walk(server_hash)) != NULL) {
      client_ptr = server_ptr->clients;
      while (client_ptr != NULL) {
	communication_ptr = client_ptr->com_list;
	while (communication_ptr != NULL) {
	  if (communication_ptr->request_length > 0 && communication_ptr->response_length > 0)
	    add_to_data_list(communication_ptr->response_length);
	  communication_ptr = communication_ptr->next;
	}
	client_ptr = client_ptr->next;
      }
    }
    pareto_model(data_list, data_list_counter);
  }
  else if (strcmp(argv[4], "4") == 0) {
    printf("Distribution of requst per host:\n");
    init_hash_walk(server_hash);
    while ((server_ptr = next_hash_walk(server_hash)) != NULL) {
      client_ptr = server_ptr->clients;
      while (client_ptr != NULL) {
	if (client_ptr->request_count > 0)
	  add_to_data_list(client_ptr->request_count);
	client_ptr = client_ptr->next;
      }
    }
    poisson_model(data_list, data_list_counter);
    //pareto_model(data_list, data_list_counter);
  }
  else if (strcmp(argv[4], "5") == 0) {
    printf("Active host count:\n");
    poisson_model(interval_active_host, time_interval_counter);
  }
  else if (strcmp(argv[4], "6") == 0) {
    printf("Host active time:\n");
    for (i = 0; i < (1 << (32 - src_subnet_mask_length)); i++)
      if  (host_start_time[i] > 0)
	add_to_data_list(host_end_time[i] - host_start_time[i]);

    exponential_model(data_list, data_list_counter);
  }

  clear_hash_table(server_hash);
  clear_hash_table(client_hash);
  clear_hash_table(communication_hash);
  coral_stop_all();
  coral_close_all();

  return 0;
}
