/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include "libcoral.h"

static void quit(int sig)
{
  coral_pkt_done = 1;
}

int main (int argc, char *argv[])
{
  int  i, j, n;
  coral_source_t *sources;
  coral_iface_t *iface;
  coral_pkt_result_t pkt_result;
  coral_interval_result_t interval_result;
  coral_pkt_buffer_t netpkt, payload;
  struct ip *ip;
  struct tcphdr *tcp;
  struct udphdr *udp;
  struct timeval interval = { 3600, 0 };
  struct timespec ts;
  uint32_t src_addr, dst_addr;
  uint16_t src_port, dst_port;
  char in_buf[64], tmp_buf[64];

  uint64_t total_packet_counter = 0;
  uint64_t total_byte_counter = 0;
  uint64_t port_byte_counter;

  uint32_t i1, i2, i3, i4, i5;

  uint32_t monitored_src, monitored_dst;
  uint32_t src_mask_length, dst_mask_length;
  char *src_subnet, *dst_subnet;
  uint64_t * port_stat;


  if (argc < 4) {
    printf("Usage: %s net1/mask1 net2/mask2 trace-file\n", argv[0]);
    printf("Example:\n");
    printf("\t%s 0.2.64.0/24 227.146.91.0/24 twoway.pcap\n", argv[0]);
    return -1;
  }


  signal(SIGINT, quit);

  //Initialize CoralReef
  coral_set_api(CORAL_API_PKT);
  coral_set_iomode(0, CORAL_RX_UNKNOWN, -1, 0); 
  coral_set_duration(0);

  //read the arguments
  printf("Processing for subnet ");
  sscanf(argv[1], "%u.%u.%u.%u/%u", &i1, &i2, &i3, &i4, &i5);
  monitored_src = ((i1 << 24) + (i2 << 16) + (i3 << 8) + i4) & (0xFFFFFFFF << (32-i5));
  src_mask_length = i5;
  printf("%u.%u.%u.%u/%u <-> ", monitored_src>>24&0xFF, monitored_src>>16&0xFF, monitored_src>>8&0xFF, monitored_src&0xFF, i5);

  sscanf(argv[2], "%u.%u.%u.%u/%u", &i1, &i2, &i3, &i4, &i5);
  monitored_dst = ((i1 << 24) + (i2 << 16) + (i3 << 8) + i4) & (0xFFFFFFFF << (32-i5));
  dst_mask_length = i5;
  printf("%u.%u.%u.%u/%u\n", monitored_dst>>24&0xFF,  monitored_dst>>16&0xFF, monitored_dst>>8&0xFF, monitored_dst&0xFF, i5);
  
  //initialize the data structure
  port_stat = (uint64_t*)malloc(65536 * 2 * sizeof(uint64_t));
  memset(port_stat, 0, sizeof(uint64_t) * 65536 * 2);

  src_subnet = (char*)malloc(1 << (32 - src_mask_length) *sizeof(char));
  memset(src_subnet, 0, sizeof(char) * (1 << (32 - src_mask_length)));

  dst_subnet = (char*)malloc(1 << (32 - dst_mask_length) *sizeof(char));
  memset(dst_subnet, 0, sizeof(char) * (1 << (32 - dst_mask_length)));


  if ((sources = coral_new_source(argv[argc-1])) == NULL)
    exit(-1);
  n = coral_open(sources);
  if (n == 0) {
    coral_diag(0, ("Could not open sources.\n"));
    exit(-1);
  }
  if (coral_start(sources) < 0) {
    fprintf(stderr, "Could not start source, aborting!\n");
    exit(-1);
  }

  coral_read_pkt_init(NULL, NULL, &interval);

  while (1) {
    if (coral_pkt_done) {
      coral_stop_all();
      exit(-1);
    }

    iface = coral_read_pkt(&pkt_result, &interval_result);
    
    // no more packets from this trace file
    if (!iface) {
      break;
    }
    
    if (pkt_result.packet) {
      if (coral_get_payload_by_layer(pkt_result.packet, &netpkt, 3) < 0)
	continue;
      
      if (netpkt.protocol != CORAL_NETPROTO_IP)
	continue;
      
      ip = (struct ip*)netpkt.buf;
      
      total_packet_counter++;

      src_addr = ntohl(ip->ip_src.s_addr);
      dst_addr = ntohl(ip->ip_dst.s_addr);


      /* For counting the traffic between src and dst subnets for different service ports */
      if ((src_addr & 0xFFFFFFFF << (32-src_mask_length)) == monitored_src && 
	  (dst_addr & 0xFFFFFFFF << (32-dst_mask_length)) == monitored_dst) {

	/* For counting active host in each subnet */
	src_subnet[src_addr & (0xFFFFFFFF >> src_mask_length)] = 1;
	dst_subnet[dst_addr & (0xFFFFFFFF >> dst_mask_length)] = 1;
	
	coral_get_payload(&netpkt, &payload);
	if (ip->ip_p == IPPROTO_TCP) {
	  tcp = (struct tcphdr*)payload.buf;
	  src_port = ntohs(tcp->th_sport);
	  dst_port = ntohs(tcp->th_dport);
	  if (src_port < 1024)
	    port_stat[src_port] += ntohs(ip->ip_len);
	  else if (dst_port < 1024)
	    port_stat[dst_port] += ntohs(ip->ip_len);
	  else
	    port_stat[1024] += ntohs(ip->ip_len);
	}
	else if (ip->ip_p == IPPROTO_UDP) {
	  udp = (struct udphdr*)payload.buf;
	  src_port = ntohs(udp->uh_sport);
	  dst_port = ntohs(udp->uh_dport);
	  if (src_port < 1024)
	    port_stat[src_port+65536] += ntohs(ip->ip_len);
	  else if (dst_port < 1024)
	    port_stat[dst_port+65536] += ntohs(ip->ip_len);
	  else
	    port_stat[1024+65536] += ntohs(ip->ip_len);
	}
	total_byte_counter += ntohs(ip->ip_len);
	total_packet_counter++;
      }       
    }
    else if (!interval_result.stats) {
    }
    else {
    }
  }
  coral_stop(sources);
  coral_close(sources);


  printf("Total packet: %"PRIu64"\tTotal byte: %"PRIu64"\n", total_packet_counter, total_byte_counter);
    
  j = 0;
  for (i = 0; i < (1 << (32 - src_mask_length)); i++)
    if (src_subnet[i] > 0)
      j++;
  printf("%u.%u.%u.%u/%u: %u active hosts\n", monitored_src>>24&0xFF, monitored_src>>16&0xFF, monitored_src>>8&0xFF, monitored_src&0xFF, src_mask_length, j);
  
  j = 0;
  for (i = 0; i < (1 << (32 - dst_mask_length)); i++)
    if (dst_subnet[i] > 0)
      j++;
  printf("%u.%u.%u.%u/%u: %u active hosts\n", monitored_dst>>24&0xFF, monitored_dst>>16&0xFF, monitored_dst>>8&0xFF, monitored_dst&0xFF, dst_mask_length, j);
  

  //print out the top five ports that are responsible for the majority of traffic
  for (i = 0; i < 5; i++) {
    n = -1;
    port_byte_counter = 0;
    for (j = 0; j < 65536 * 2; j++)
      if (port_stat[j] > port_byte_counter) {
	n = j;
	port_byte_counter = port_stat[j];
      }
    if (port_byte_counter == 0)
      break;
    if (n < 65536)
      printf("<TCP%i %f\%> ", n, port_byte_counter * 100.0 / total_byte_counter);
    else
      printf("<UDP%i %f\%> ", n-65536, port_byte_counter * 100.0 / total_byte_counter);
    port_stat[n] = 0;
  }
  printf("\n");
  
  coral_stop_all();
  coral_close_all();

  return 0;
}
