/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include "libcoral.h"

#define MAX_SOURCES 240
#define MAX_SRC 10
#define MAX_DST 200

static void quit(int sig)
{
  coral_pkt_done = 1;
}

int main (int argc, char *argv[])
{
  int  i, j, n;

  coral_source_t *sources;
  coral_iface_t *iface;
  coral_pkt_result_t pkt_result;
  coral_interval_result_t interval_result;
  coral_pkt_buffer_t netpkt, payload;

  struct ip *ip;
  struct tcphdr *tcp;
  struct udphdr *udp;
  struct timeval interval = { 3600, 0 };
  struct timespec ts;
  uint32_t src_addr, dst_addr;
  uint16_t src_port, dst_port;
  char in_buf[64], tmp_buf[64];
  int error = 0;

  uint64_t total_packet_counter = 0;
  uint64_t total_byte_counter = 0;
  uint64_t subnet_byte_counter;

  uint32_t i1, i2, i3, i4;

  uint32_t monitored_subnet; //IPv4 address of the monitored subnet
  uint32_t mask_length; //mask length in bit
  int send_receive; //switch to counte incoming or outgoing traffic
  uint64_t *subnet_stat;

  if (argc <= 3 || 
      (strcmp(argv[1], "send") != 0 && strcmp(argv[1], "receive") != 0) ||
      strchr(argv[2], '/') != argv[2]) {
    printf("Usage: %s [send|receive] mask-length [subnet] trace-file\n", argv[0]);
    printf("Example:\n");
    printf("\t%s send /24 twoway.pcap\n", argv[0]);
    printf("\t%s receiv /24 0.2.64.5 twoway.pcap\n", argv[0]);
    return -1;
  }

  send_receive = strcmp(argv[1], "send");
  mask_length = atoi(argv[2]+1);
  if (argc >= 5) {
    sscanf(argv[3], "%u.%u.%u.%u", &i1, &i2, &i3, &i4);
    monitored_subnet = ((i1 << 24) + (i2 << 16) + (i3 << 8) + i4) & (0xFFFFFFFF << (32 - mask_length));
  }
  else
    monitored_subnet = 0xFFFFFFFF;


  //Initialize CoralReef
  signal(SIGINT, quit);
  coral_set_api(CORAL_API_PKT);
  coral_set_iomode(0, CORAL_RX_UNKNOWN, -1, 0); 
  coral_set_duration(0);

  //nintialize the statistics data structure
  subnet_stat = (uint64_t*)malloc((1 << mask_length) * sizeof(uint64_t));
  memset(subnet_stat, 0, sizeof(uint64_t) * (1 << mask_length));


  if ((sources = coral_new_source(argv[argc-1])) == NULL)
    exit(-1);
  
  if (coral_open(sources) == 0) {
    coral_diag(0, ("Could not open sources.\n"));
    exit(-1);
  }

  if (coral_start(sources) < 0) {
    fprintf(stderr, "Could not start source, aborting!\n");
    exit(-1);
  }
  
  coral_read_pkt_init(NULL, NULL, &interval);

  error = 1;
  while (1) {
    if (coral_pkt_done) {
      coral_stop_all();
      exit(-1);
    }

    iface = coral_read_pkt(&pkt_result, &interval_result);

    // no more packets from this trace file
    if (!iface) {
      break;
    }

    if (pkt_result.packet) {
      if (coral_get_payload_by_layer(pkt_result.packet, &netpkt, 3) < 0)
	continue;
       
      if (netpkt.protocol != CORAL_NETPROTO_IP)
	continue;
	
      ip = (struct ip*)netpkt.buf;
      
      total_packet_counter++;

      src_addr = ntohl(ip->ip_src.s_addr);
      dst_addr = ntohl(ip->ip_dst.s_addr);

      if (monitored_subnet != 0xFFFFFFFF) {
	if (send_receive == 0 && (dst_addr & (0xFFFFFFFF << (32 - mask_length))) == monitored_subnet) {
	  total_byte_counter += ntohs(ip->ip_len);
	  subnet_stat[src_addr >> (32-mask_length)] += ntohs(ip->ip_len);
	}
	else if (send_receive != 0 && (src_addr & (0xFFFFFFFF << (32 - mask_length))) == monitored_subnet) {
	  total_byte_counter += ntohs(ip->ip_len);
	  subnet_stat[dst_addr >> (32-mask_length)] += ntohs(ip->ip_len);
	}
      }
      else {
	total_byte_counter += ntohs(ip->ip_len);
	if (send_receive == 0)
	  subnet_stat[src_addr >> (32-mask_length)] += ntohs(ip->ip_len);
	else
	  subnet_stat[dst_addr >> (32-mask_length)] += ntohs(ip->ip_len);
      }
    }
    else if (!interval_result.stats) {
    }
    else {
    }
  }

  coral_stop(sources);
  coral_close(sources);

  printf("Total packets: %"PRIu64"\tTotal bytes: %"PRIu64"\n", total_packet_counter, total_byte_counter);

  //print out the statistics of each subnet
  for (i = 0; i < (1 << mask_length); i++) {
    subnet_byte_counter = 0;
    n = 0;
    for (j = 0; j < (1 << mask_length); j++)
      if (subnet_stat[j] > subnet_byte_counter) {
	n = j;
	subnet_byte_counter = subnet_stat[j];
      }
    if (subnet_byte_counter == 0)
      break;
    subnet_stat[n] = 0;
    monitored_subnet = n << (32 - mask_length);
    printf("%u.%u.%u.%u/%u\t%f\t%"PRIu64"\n", monitored_subnet>>24&0xFF, monitored_subnet>>16&0xFF, monitored_subnet>>8&0xFF, monitored_subnet&0xFF, mask_length, subnet_byte_counter*100.0/total_byte_counter, subnet_byte_counter);
  }

  coral_stop_all();
  coral_close_all();
  
  return 0;
}
